#!/usr/bin/python
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import json
import sys

from ..module_utils import bogusmu  # pylint: disable=relative-beyond-top-level


def main():
    raise Exception('should never get here')


if __name__ == '__main__':
    main()
